"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import functools
from typing import List

import pytest
from unittest.mock import Mock

from proton.vpn.session.servers import LogicalServer, ServerFeatureEnum
from proton.vpn.session.servers.logicals import (
    sort_servers_alphabetically_by_country_and_server_name,
    sort_servers_by_country_and_location_and_enabled_and_load,
    ServerList
)


def _compact_features(features: List[ServerFeatureEnum]) -> ServerFeatureEnum:
    return functools.reduce(lambda f1, f2: f1 | f2, features, 0)


@pytest.fixture(name="api_response")
def fixture_api_response() -> str:
    return {
        "Code": 1000,
        "LogicalServers": [
            {
                "ID": 1,
                "Name": "JP#10",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 15.0,  # AR#9 has better score (lower is better)
                "Tier": 2,
                "ExitCountry": "JP",
                "City": "Tokyo",
                "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.STREAMING
            },
            {
                "ID": 2,
                "Name": "AR#11",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 1.0,  # Even though it has a better score than CH#9,
                "Tier": 3,     # it's not in the user tier (2).
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.STREAMING
            },
            {
                "ID": 3,
                "Name": "AR#13",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 2.0,  # Even though it has a better score than CH#9,
                "Tier": 3,     # it's not in the user tier (2).
                "ExitCountry": "AR",
                "City": "Rosario",
                "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.STREAMING | ServerFeatureEnum.IPV6
            },
            {
                "ID": 4,
                "Name": "AR#14",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 3.0,  # Even though it has a better score than CH#9,
                "Tier": 3,     # it's not in the user tier (2).
                "ExitCountry": "AR",
                "City": "Rosario",
                "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.IPV6
            },
            {
                "ID": 5,
                "Name": "AR#9",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 10.0,  # Fastest server in the  user tier (2)
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Rosario"
            },
            {
                "ID": 6,
                "Name": "CH#18-TOR",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 7.0,                       # Even though it has a better score than AR#9,
                "Features": ServerFeatureEnum.TOR,  # TOR servers should be ignored.
                "Tier": 2,
                "ExitCountry": "CH",
                "City": "Wuhan"
            },
            {
                "ID": 7,
                "Name": "CH-US#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Score": 8.0,                               # Even though it has a better score than AR#9,
                "Features": ServerFeatureEnum.SECURE_CORE,  # secure core servers should be ignored.
                "Tier": 2,
                "ExitCountry": "CH",
                "City": "Beijing"
            },
            {
                "ID": 8,
                "Name": "JP#1",
                "Score": 9.0,  # Even though it has a better score than AR#9,
                "Status": 0,   # this server is not enabled.
                "Servers": [{"Status": 0}],
                "Tier": 2,
                "ExitCountry": "JP",
                "City": "Osaka",
            },
        ]
    }


def test_server_list_get_fastest(api_response: str):
    server_list = ServerList(
        user_tier=2,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    fastest = server_list.get_fastest()
    assert fastest.name == "AR#9"


def test_server_list_get_fastest_in_country(api_response: str):
    server_list = ServerList(
        user_tier=3,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    fastest = server_list.get_fastest_in_country("AR")
    assert fastest.name == "AR#11"


def test_server_list_get_fastest_in_city(api_response: str):
    server_list = ServerList(
        user_tier=3,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    fastest = server_list.get_fastest_in_city("Rosario")
    assert fastest.name == "AR#13"


def test_server_list_get_available_servers(api_response: str):
    server_list = ServerList(
        user_tier=2,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    available_servers_generator = \
        ServerList.get_available_servers(server_list.logicals, user_tier=2)

    for server in available_servers_generator:
        assert server.tier <= 2 and server.enabled


@pytest.mark.parametrize("features_requested, features_excluded", [
    (ServerFeatureEnum.P2P, 0),
    (ServerFeatureEnum.IPV6 | ServerFeatureEnum.STREAMING, 0),
    (ServerFeatureEnum.P2P, ServerFeatureEnum.TOR | ServerFeatureEnum.IPV6),
    (0, 0),
    (ServerFeatureEnum.TOR, ServerFeatureEnum.TOR),
])
def test_server_list_get_servers_with_features(
    api_response: str,
    features_requested: ServerFeatureEnum,
    features_excluded: ServerFeatureEnum
):
    server_list = ServerList(
        user_tier=2,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    servers_with_features_generator = \
        ServerList.get_servers_with_features(
            server_list.logicals,
            features_requested,
            features_excluded
        )

    servers_with_features = list(servers_with_features_generator)

    # test for empty server list when requested and excluded features have an intersection
    if features_requested & features_excluded != 0:
        assert len(servers_with_features) == 0

    for server in servers_with_features:
        assert _compact_features(server.features) & features_excluded == 0 \
            and _compact_features(server.features) & features_requested == features_requested


def test_server_list_get_country_servers(api_response: str):
    server_list = ServerList(
        user_tier=2,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    country_servers_generator = \
        ServerList.get_servers_in_city(server_list.logicals, "AR")

    for server in country_servers_generator:
        assert server.exit_country == "AR"


def test_server_list_get_city_servers(api_response: str):
    server_list = ServerList(
        user_tier=2,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    city_servers_generator = \
        ServerList.get_servers_in_country_code(server_list.logicals, "rosario")

    for server in city_servers_generator:
        assert server.city == "rosario"


def test_server_list_get_fastest_server(api_response: str):
    server_list = ServerList(
        user_tier=2,
        logicals=[LogicalServer(ls) for ls in api_response["LogicalServers"]]
    )

    fastest_server = ServerList.get_fastest_server(server_list.logicals)

    assert fastest_server.name == "AR#11"


def test_sort_servers_alphabetically_by_country_and_server_name():
    api_response = {
        "Code": 1000,
        "LogicalServers": [
            {
                "ID": 2,
                "Name": "AR#10",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "ExitCountry": "AR",
            },
            {
                "ID": 1,
                "Name": "JP-FREE#10",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "ExitCountry": "JP",
            },
            {
                "ID": 3,
                "Name": "AR#9",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "ExitCountry": "AR",
            },
            {
                "ID": 5,
                "Name": "Random Name",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "ExitCountry": "JP",
            },
            {
                "ID": 4,
                "Name": "JP#9",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "ExitCountry": "JP",

            },
        ]
    }

    logicals = [LogicalServer(server_dict) for server_dict in api_response["LogicalServers"]]
    logicals.sort(key=sort_servers_alphabetically_by_country_and_server_name)

    expected_server_name_order = ["AR#9", "AR#10", "JP#9", "JP-FREE#10", "Random Name"]
    actual_server_name_order = [server.name for server in logicals]
    assert actual_server_name_order == expected_server_name_order


def test_sort_servers_by_country_and_city_and_enabled_and_load():
    api_response = {
        "Code": 1000,
        "LogicalServers": [
            {
                "ID": 1,
                "Name": "AR#10",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Load": 50,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
            },
            {
                "ID": 2,
                "Name": "AR#9",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Load": 30,  # Lower load, should come before AR#10 in same city
                "ExitCountry": "AR",
                "City": "Buenos Aires",
            },
            {
                "ID": 3,
                "Name": "AR#5",
                "Status": 0,  # Disabled, should come after enabled servers
                "Servers": [{"Status": 0}],
                "Load": 20,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
            },
            {
                "ID": 4,
                "Name": "AR#15",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Load": 20,  # Lowest load in Rosario
                "ExitCountry": "AR",
                "City": "Rosario",
            },
            {
                "ID": 5,
                "Name": "JP#9",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Load": 40,
                "ExitCountry": "JP",
                "City": "Tokyo",
            },
            {
                "ID": 6,
                "Name": "JP#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Load": 25,  # Lower load than JP#9
                "ExitCountry": "JP",
                "City": "Tokyo",
            },
            {
                "ID": 7,
                "Name": "JP#10",
                "Status": 0,  # Disabled
                "Servers": [{"Status": 0}],
                "Load": 10,  # Even lower load, but disabled
                "ExitCountry": "JP",
                "City": "Tokyo",
            },
            {
                "ID": 8,
                "Name": "JP#5",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Load": 35,
                "ExitCountry": "JP",
                "City": "Osaka",  # Different city, should come after Tokyo
            },
        ]
    }

    logicals = [LogicalServer(server_dict) for server_dict in api_response["LogicalServers"]]
    logicals.sort(key=sort_servers_by_country_and_location_and_enabled_and_load)

    # Expected order:
    # Enabled servers should come first (True sorts after False, but the function should invert this)
    # 1. Argentina (AR) - Buenos Aires - Enabled first (by load: 30 < 50), then disabled
    #    - AR#9 (enabled, load 30)
    #    - AR#10 (enabled, load 50)
    #    - AR#5 (disabled, load 20)
    # 2. Argentina (AR) - Rosario
    #    - AR#15 (enabled, load 20)
    # 3. Japan (JP) - Osaka
    #    - JP#5 (enabled, load 35)
    # 4. Japan (JP) - Tokyo - Enabled first (by load: 25 < 40), then disabled
    #    - JP#1 (enabled, load 25)
    #    - JP#9 (enabled, load 40)
    #    - JP#10 (disabled, load 10)
    expected_server_name_order = [
        "AR#9",      # Argentina, Buenos Aires, enabled, load 30
        "AR#10",     # Argentina, Buenos Aires, enabled, load 50
        "AR#5",      # Argentina, Buenos Aires, disabled, load 20
        "AR#15",     # Argentina, Rosario, enabled, load 20
        "JP#5",      # Japan, Osaka, enabled, load 35
        "JP#1",      # Japan, Tokyo, enabled, load 25
        "JP#9",      # Japan, Tokyo, enabled, load 40
        "JP#10",     # Japan, Tokyo, disabled, load 10
    ]
    actual_server_name_order = [server.name for server in logicals]
    assert actual_server_name_order == expected_server_name_order



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future
from unittest.mock import Mock, patch, AsyncMock, DEFAULT

import gi
from proton.vpn.connection.persistence import ConnectionParameters
from proton.vpn.session.servers.types import TierEnum

gi.require_version("NM", "1.0")  # noqa: required before importing NM module
from gi.repository import NM, GLib

import pytest

from tests.networkmanager.core.boilerplate import VPNServer, VPNCredentials, Settings
from proton.vpn.backend.networkmanager.core import LinuxNetworkManager
from proton.vpn.connection.events import EventContext
from proton.vpn.connection import states
from proton.vpn.connection import events
from collections import namedtuple

OpenVPNPorts = namedtuple("OpenVPNPorts", "udp tcp")


class LinuxNetworkManagerProtocol(LinuxNetworkManager):
    """Dummy protocol just to unit test the base LinuxNetworkManager class."""
    protocol = "Dummy protocol"

    @classmethod
    def get_protocol_group(cls) -> str:
        return "dummy"

    def __init__(self, *args, connection_persistence=None, **kwargs):
        # Make sure we don't trigger connection persistence nor the kill switch.
        connection_persistence = connection_persistence or Mock()

        super().__init__(*args, connection_persistence=connection_persistence, user_tier=TierEnum.PLUS,
                         **kwargs)

    def setup(self):
        # to be mocked in tests
        pass


@pytest.fixture
def nm_client_mock():
    return Mock()


def create_nm_protocol(nm_client_mock):
    return LinuxNetworkManagerProtocol(
        VPNServer(
                openvpn_ports=OpenVPNPorts([00], [00])
        ), VPNCredentials(), Settings(), nm_client=nm_client_mock
    )


@pytest.mark.asyncio
@patch("proton.vpn.backend.networkmanager.core.networkmanager.tcpcheck")
async def test_start(tcpcheck_patch, nm_client_mock):
    # Mock successful TCP connection check.
    tcpcheck_patch.is_any_port_reachable = AsyncMock()

    nm_protocol = create_nm_protocol(nm_client_mock)

    with patch.object(nm_protocol, "setup") as setup_mock:
        start_connection_future = Future()
        nm_client_mock.start_connection_async.return_value = start_connection_future
        connection_mock = setup_mock.return_value.result()
        start_connection_future.set_result(connection_mock)

        await nm_protocol.start()

        setup_mock.assert_called_once()

    nm_client_mock.start_connection_async.assert_called_once_with(connection_mock)

    # Assert that once the connection has been activated, the expected callback
    # is hooked to monitor vpn connection state changes.
    connection_mock.connect.assert_called_once_with(
        "vpn-state-changed",
        nm_protocol._on_state_changed
    )


@pytest.mark.asyncio
@patch("proton.vpn.backend.networkmanager.core.networkmanager.tcpcheck")
async def test_start_generates_timeout_event_when_the_tcp_connection_check_fails(
        tcpcheck_patch, nm_client_mock
):
    # Mock failed TCP connection check.
    tcpcheck_patch.is_any_port_reachable = AsyncMock(return_value=False)

    connection_subscriber = Mock()
    nm_protocol = create_nm_protocol(nm_client_mock)
    nm_protocol.register(connection_subscriber)
    with patch.object(nm_protocol, "setup") as setup_mock:
        await nm_protocol.start()

        setup_mock.assert_not_called()

    connection_subscriber.assert_called_once()

    generated_event = connection_subscriber.call_args.kwargs["event"]
    assert isinstance(generated_event, events.Timeout)


@pytest.mark.asyncio
@patch("proton.vpn.backend.networkmanager.core.networkmanager.tcpcheck")
async def test_start_generates_tunnel_setup_failed_event_on_connection_setup_errors(
        tcpcheck_patch, nm_client_mock
):
    nm_protocol = create_nm_protocol(nm_client_mock)

    # Mock successful TCP connection check.
    tcpcheck_patch.is_any_port_reachable = AsyncMock(return_value=True)

    with patch.object(nm_protocol, "setup") as setup_mock:
        # Mock error on connection setup.
        setup_connection_future = Future()
        setup_connection_future.set_exception(GLib.GError)
        setup_mock.return_value = setup_connection_future

        connection_subscriber = Mock()
        nm_protocol.register(connection_subscriber)
        await nm_protocol.start()

        setup_mock.assert_called()

    connection_subscriber.assert_called_once()

    generated_event = connection_subscriber.call_args.kwargs["event"]
    assert isinstance(generated_event, events.TunnelSetupFailed)


@pytest.mark.asyncio
@patch("proton.vpn.backend.networkmanager.core.networkmanager.tcpcheck")
async def test_start_generates_tunnel_setup_failed_event_on_connection_activation_errors_and_removes_connection(
        tcpcheck_patch, nm_client_mock
):
    nm_protocol = create_nm_protocol(nm_client_mock)

    # Mock successful TCP connection check.
    tcpcheck_patch.is_any_port_reachable = AsyncMock(return_value=True)

    with patch.multiple(nm_protocol, setup=DEFAULT, remove_connection=DEFAULT) as mocks:
        # Mock successful connection setup.
        connection = Mock()
        setup_connection_future = Future()
        setup_connection_future.set_result(connection)
        mocks["setup"].return_value = setup_connection_future

        # Mock error on connection activation.
        start_connection_future = Future()
        start_connection_future.set_exception(GLib.GError)
        nm_client_mock.start_connection_async.return_value = start_connection_future

        connection_subscriber = Mock()
        nm_protocol.register(connection_subscriber)
        await nm_protocol.start()

        nm_client_mock.start_connection_async.assert_called_once_with(connection)
        connection_subscriber.assert_called_once()

        generated_event = connection_subscriber.call_args.kwargs["event"]
        assert isinstance(generated_event, events.TunnelSetupFailed)

        mocks["remove_connection"].assert_called_once()


@pytest.mark.asyncio
async def test_remove_connection(nm_client_mock):
    nm_protocol = create_nm_protocol(nm_client_mock)
    connection_mock = Mock()
    await nm_protocol.remove_connection(connection_mock)
    nm_client_mock.remove_connection_async.assert_called_once_with(connection_mock)
    assert nm_protocol._unique_id is None


@pytest.mark.asyncio
async def test_stop_connection_removes_connection(nm_client_mock):
    nm_protocol = create_nm_protocol(nm_client_mock)
    with patch.object(nm_protocol, "remove_connection"):
        connection = Mock()
        await nm_protocol.stop(connection)

        nm_protocol.remove_connection.assert_called_once_with(connection)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "active_nm_connection, inactive_nm_connection, expected_state",
    [
        (
                Mock(),
                None,
                states.Connected,  # When there is an active connection the initial state is connected.
        ),
        (
                None,
                None,
                states.Disconnected  # When there is not a connection, the initial state is disconnected.
        ),
        (
                None,
                Mock(),
                states.Error  # When there is an inactive connection, the initial state is Error.
        ),
    ]
)
async def test_initialize_persisted_connection_determines_initial_connection_state(
        active_nm_connection, inactive_nm_connection, expected_state
):
    nm_client_mock = Mock()
    nm_client_mock.get_active_connection.return_value = active_nm_connection
    nm_client_mock.get_connection.return_value = inactive_nm_connection

    # The VPNConnection constructor calls `_initialize_persisted_connection`
    # when `connection_id` is provided.
    nm_protocol = LinuxNetworkManagerProtocol(
        server=None,
        credentials=None,
        settings=None,
        connection_id="connection_id",
        nm_client=nm_client_mock
    )

    assert isinstance(nm_protocol.initial_state, expected_state)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock, patch

import gi

gi.require_version("NM", "1.0")  # noqa: required before importing NM module
from gi.repository import NM

import pytest

from proton.vpn.backend.networkmanager.protocol.openvpn.openvpn \
        import OpenVPN, OpenVPNTCP
from proton.vpn.connection import events
from collections import namedtuple
from proton.vpn.session.servers.types import TierEnum

from boilerplate import (MockVpnCredentials, MockSettings, vpn_server)

OpenVPNPorts = namedtuple("OpenVPNPorts", "udp tcp")


@pytest.fixture
def nm_client_mock():
    return Mock()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "state, reason, expected_event",
    [
        (
                NM.VpnConnectionState.ACTIVATED,
                NM.VpnConnectionStateReason.NONE,
                events.Connected
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.CONNECT_TIMEOUT,
                events.Timeout
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.SERVICE_START_TIMEOUT,
                events.Timeout
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.NO_SECRETS,
                events.AuthDenied
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.LOGIN_FAILED,
                events.AuthDenied
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.IP_CONFIG_INVALID,
                events.UnexpectedError
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.SERVICE_STOPPED,
                events.UnexpectedError
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.CONNECTION_REMOVED,
                events.UnexpectedError
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.SERVICE_START_FAILED,
                events.UnexpectedError
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.UNKNOWN,
                events.UnexpectedError
        ),
        (
                NM.VpnConnectionState.FAILED,
                NM.VpnConnectionStateReason.NONE,
                events.UnexpectedError
        ),
        (
                NM.VpnConnectionState.DISCONNECTED,
                NM.VpnConnectionStateReason.DEVICE_DISCONNECTED,
                events.DeviceDisconnected
        ),
        (
                NM.VpnConnectionState.DISCONNECTED,
                NM.VpnConnectionStateReason.USER_DISCONNECTED,
                events.Disconnected
        ),
        (
                NM.VpnConnectionState.DISCONNECTED,
                NM.VpnConnectionStateReason.NONE,
                events.UnexpectedError
        ),
    ]
)
@patch("proton.vpn.backend.networkmanager.protocol.openvpn.openvpn.OpenVPN._notify_subscribers_threadsafe")
async def test_on_state_changed(_notify_subscribers_threadsafe, nm_client_mock,
                                vpn_server, state, reason, expected_event):
    _notify_subscribers_threadsafe.return_value = None

    nm_protocol = OpenVPNTCP(
        vpn_server, MockVpnCredentials(), MockSettings(),
        nm_client=nm_client_mock, user_tier=TierEnum.PLUS
    )
    nm_protocol._on_state_changed(None, state, reason)

    # assert that the OpenVPN._notify_subscribers method was called with the
    # expected event
    _notify_subscribers_threadsafe.assert_called_once()
    assert isinstance(_notify_subscribers_threadsafe.call_args.args[0],
                      expected_event)



"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import asdict
import pytest
from proton.vpn.session.dataclasses import VPNSessions, APIVPNSession


@pytest.fixture
def vpnsession_data():
    return {
        "SessionID": "session1",
        "ExitIP": "2.2.2.1",
        "Protocol": "openvpn-tcp",
    }


def test_vpnsession_deserializes_expected_dict_keys(vpnsession_data):
    vpnsession = APIVPNSession.from_dict(vpnsession_data)

    assert asdict(vpnsession) == vpnsession_data


def test_vpnsession_deserialize_should_not_crash_with_unexpected_dict_keys(vpnsession_data):
    vpnsession_data["unexpected_keyword"] = "keyword and data"

    APIVPNSession.from_dict(vpnsession_data)


@pytest.fixture
def vpnsessions_data():
    return {
        "Sessions": [
            {
                "SessionID": "session1",
                "ExitIP": "2.2.2.1",
                "Protocol": "openvpn-tcp",
            },
            {
                "SessionID": "session2",
                "ExitIP": "2.2.2.3",
                "Protocol": "openvpn-udp",
            },
            {
                "SessionID": "session3",
                "ExitIP": "2.2.2.53",
                "Protocol": "wireguard",
            }
        ]
    }


def test_vpnsessions_deserializes_expected_dict_keys(vpnsessions_data):
    vpnsessions = VPNSessions.from_dict(vpnsessions_data)

    assert asdict(vpnsessions) == vpnsessions_data


def test_vpnsessions_deserialize_should_not_crash_with_unexpected_dict_keys(vpnsessions_data):
    vpnsessions_data["unexpected_keyword"] = "keyword and data"

    VPNSessions.from_dict(vpnsessions_data)





"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytest
from proton.vpn.session.servers.types import LogicalServer, ServerFeatureEnum
from proton.vpn.session.dataclasses.servers.country import Location, Country, ServerAnalysis


COUNTRY_CODE = "AR"
COUNTRY_NAME = "Argentina"
CITIES = ["Buenos Aires", "Rosario"]
ROSARIO_CITY_FEATURES = {
    ServerFeatureEnum.P2P, ServerFeatureEnum.STREAMING, ServerFeatureEnum.IPV6
}


@pytest.fixture()
def servers_raw() -> list[dict]:
    return [
                {
            "ID": 2,
            "Name": "AR#11",
            "Status": 1,
            "Servers": [{"Status": 1}],
            "Score": 1.0,  # Even though it has a better score than CH#9,
            "Tier": 2,     # it's not in the user tier (2).
            "ExitCountry": "AR",
            "City": "Buenos Aires",
            "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.STREAMING
        },
        {
            "ID": 3,
            "Name": "AR#13",
            "Status": 1,
            "Servers": [{"Status": 1}],
            "Score": 2.0,  # Even though it has a better score than CH#9,
            "Tier": 2,     # it's not in the user tier (2).
            "ExitCountry": "AR",
            "City": "Rosario",
            "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.STREAMING | ServerFeatureEnum.IPV6
        },
        {
            "ID": 4,
            "Name": "AR#14",
            "Status": 1,
            "Servers": [{"Status": 1}],
            "Score": 2.0,  # Even though it has a better score than CH#9,
            "Tier": 2,     # it's not in the user tier (2).
            "ExitCountry": "AR",
            "City": "Rosario",
            "Features": ServerFeatureEnum.P2P
        }
    ]


@pytest.fixture()
def non_free_logical_servers(servers_raw) -> list[LogicalServer]:
    return [
        LogicalServer(servers_raw[0]),
        LogicalServer(servers_raw[1]),
        LogicalServer(servers_raw[2])
    ]


@pytest.fixture()
def free_logical_servers(servers_raw) -> list[LogicalServer]:
    servers_raw[0]["Tier"] = 0
    servers_raw[1]["Tier"] = 0
    servers_raw[2]["Tier"] = 0

    return [
        LogicalServer(servers_raw[0]),
        LogicalServer(servers_raw[1]),
        LogicalServer(servers_raw[2])
    ]


@pytest.fixture()
def mixed_free_and_non_logical_servers(servers_raw) -> list[LogicalServer]:
    servers_raw[1]["Tier"] = 0

    return [
        LogicalServer(servers_raw[0]),
        LogicalServer(servers_raw[1]),
        LogicalServer(servers_raw[2])
    ]


class TestCountry:
    def test_name_is_correctly_returned_when_passing_country_code(self):
        country = Country(COUNTRY_CODE, [])
        assert country.name == COUNTRY_NAME

    def test_locations_are_grouped(self, non_free_logical_servers):
        country = Country(COUNTRY_CODE, non_free_logical_servers, group_by_location=True)

        locations = country.locations

        assert locations[0].name == CITIES[0]
        assert locations[1].name == CITIES[1]
        assert len(locations) == len(CITIES)

    def test_locations_are_grouped_by_state_when_state_is_present(self):
        """When servers have a State field it takes priority over City for grouping."""
        servers = [
            LogicalServer({
                "ID": 1, "Name": "US#1", "Status": 1, "Servers": [{"Status": 1}],
                "Tier": 2, "ExitCountry": "US",
                "State": "California", "City": "Los Angeles", "Features": 0,
            }),
            LogicalServer({
                "ID": 2, "Name": "US#2", "Status": 1, "Servers": [{"Status": 1}],
                "Tier": 2, "ExitCountry": "US",
                "State": "California", "City": "San Francisco", "Features": 0,
            }),
            LogicalServer({
                "ID": 3, "Name": "US#3", "Status": 1, "Servers": [{"Status": 1}],
                "Tier": 2, "ExitCountry": "US",
                "State": "Texas", "City": "Houston", "Features": 0,
            }),
        ]
        country = Country("US", servers, group_by_location=True)

        locations = country.locations

        assert len(locations) == 2
        assert locations[0].name == "California"
        assert locations[1].name == "Texas"
        assert len(locations[0].servers) == 2
        assert len(locations[1].servers) == 1


class TestServerAnalysis:
    def test_analyze_servers_returns_under_maintenance_true_when_all_servers_are_under_maintenance(self):
        # Create servers that are all under maintenance (Status=0 or no enabled physical servers)
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 0,  # Logical server disabled
                "Servers": [{"Status": 0}],  # Physical server disabled
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 0,
                "Servers": [{"Status": 0}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        assert analysis.under_maintenance is True

    def test_analyze_servers_returns_under_maintenance_false_when_some_servers_are_not_under_maintenance(self):
        # Create servers where at least one is enabled (not under maintenance)
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 0,  # Logical server disabled
                "Servers": [{"Status": 0}],  # Physical server disabled
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 1,  # Logical server enabled
                "Servers": [{"Status": 1}],  # Physical server enabled
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        assert analysis.under_maintenance is False

    def test_analyze_servers_returns_smart_routing_when_any_server_has_smart_routing(self):
        # Note: Based on implementation, smart_routing is True only when ALL servers have smart_routing
        # The test name says "any" but the logic requires "all"
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0,
                "HostCountry": "US"  # Smart routing enabled
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0,
                "HostCountry": "US"  # Smart routing enabled
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        assert analysis.smart_routing is True

    def test_analyze_servers_returns_smart_routing_false_when_not_all_servers_have_smart_routing(self):
        # Create servers where at least one doesn't have smart routing
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0,
                "HostCountry": "US"  # Smart routing enabled
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
                # No HostCountry - smart routing not enabled
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        assert analysis.smart_routing is False

    def test_analyze_servers_returns_free_when_any_server_is_free(self):
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,  # Not free
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 0,  # Free tier
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        assert analysis.free is True

    def test_analyze_servers_returns_free_false_when_no_servers_are_free(self):
        # Create servers where none are free (all have tier > 0)
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,  # Not free
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,  # Not free
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": 0
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        assert analysis.free is False

    def test_analyze_servers_groups_features_from_all_servers(self):
        servers_data = [
            {
                "ID": 1,
                "Name": "AR#1",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Buenos Aires",
                "Features": ServerFeatureEnum.P2P | ServerFeatureEnum.STREAMING
            },
            {
                "ID": 2,
                "Name": "AR#2",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Rosario",
                "Features": ServerFeatureEnum.IPV6 | ServerFeatureEnum.TOR
            },
            {
                "ID": 3,
                "Name": "AR#3",
                "Status": 1,
                "Servers": [{"Status": 1}],
                "Tier": 2,
                "ExitCountry": "AR",
                "City": "Rosario",
                "Features": ServerFeatureEnum.P2P  # Overlapping feature
            }
        ]
        servers = [LogicalServer(data) for data in servers_data]

        analysis = ServerAnalysis.analyze_servers(servers)

        # Features should be a set containing all unique features from all servers
        expected_features = {
            ServerFeatureEnum.P2P,
            ServerFeatureEnum.STREAMING,
            ServerFeatureEnum.IPV6,
            ServerFeatureEnum.TOR
        }
        assert analysis.features == expected_features




import time
from unittest.mock import AsyncMock

import pytest

from proton.vpn.core.refresher.scheduler import Scheduler


async def dummy():
    pass


@pytest.mark.asyncio
async def test_start_runs_tasks_ready_to_fire_periodically():
    scheduler = Scheduler(check_interval_in_ms=10)

    task_1 = AsyncMock()
    async def task_1_wrapper():
        await task_1()

    scheduler.run_after(0, task_1_wrapper)

    task_2 = AsyncMock()
    async def run_task_2_and_shutdown():
        await task_2()
        await scheduler.stop()  # stop the scheduler after the second task is executed.

    in_100_ms = time.time() + 0.1
    scheduler.run_at(in_100_ms, run_task_2_and_shutdown)

    scheduler.start()

    await scheduler.wait_for_shutdown()
    task_1.assert_called_once()
    task_2.assert_called_once()
    assert len(scheduler.task_list) == 0


@pytest.mark.asyncio
async def test_run_task_ready_to_fire_only_runs_tasks_with_expired_timestamps():
    scheduler = Scheduler()

    # should run since the delay is 0 seconds.
    scheduler.run_after(delay_in_seconds=0, async_function=dummy)
    # should not run yet since the delay is 30 seconds.
    scheduler.run_after(delay_in_seconds=30, async_function=dummy)
    scheduler.run_tasks_ready_to_fire()

    assert scheduler.number_of_remaining_tasks == 1


@pytest.mark.asyncio
async def test_stop_empties_task_list():
    scheduler = Scheduler()
    scheduler.start()

    scheduler.run_after(delay_in_seconds=30, async_function=dummy)
    await scheduler.stop()

    assert not scheduler.is_started
    assert len(scheduler.task_list) == 0


def test_run_at_schedules_new_task():
    scheduler = Scheduler()

    task_id = scheduler.run_at(timestamp=time.time() + 10, async_function=dummy)

    scheduler.task_list[0].id == task_id


def test_cancel_task_removes_task_from_task_list():
    scheduler = Scheduler()

    task_id = scheduler.run_after(0, dummy)
    scheduler.cancel_task(task_id)

    assert scheduler.number_of_remaining_tasks == 0



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn.session.dataclasses.notifications.nps_survey_response import NPSSurveyResponse


def test_nps_survey_response_clips_score_below_minimum():
    response = NPSSurveyResponse(user_score=NPSSurveyResponse.SCORE_MIN - 1)
    assert response.user_score == NPSSurveyResponse.SCORE_MIN


def test_nps_survey_response_clips_score_above_maximum():
    response = NPSSurveyResponse(user_score=NPSSurveyResponse.SCORE_MAX + 1)
    assert response.user_score == NPSSurveyResponse.SCORE_MAX


def test_nps_survey_response_accepts_score_at_minimum_boundary():
    response = NPSSurveyResponse(user_score=NPSSurveyResponse.SCORE_MIN)
    assert response.user_score == NPSSurveyResponse.SCORE_MIN


def test_nps_survey_response_accepts_score_at_maximum_boundary():
    response = NPSSurveyResponse(user_score=NPSSurveyResponse.SCORE_MAX)
    assert response.user_score == NPSSurveyResponse.SCORE_MAX


def test_nps_survey_response_truncates_comment_to_max_length():
    long_comment = "x" * (NPSSurveyResponse.COMMENT_CHAR_MAX_LENGTH + 50)
    response = NPSSurveyResponse(user_comments=long_comment)
    assert len(response.user_comments) == NPSSurveyResponse.COMMENT_CHAR_MAX_LENGTH


def test_nps_survey_response_does_not_truncate_short_comment():
    short_comment = "x" * (NPSSurveyResponse.COMMENT_CHAR_MAX_LENGTH - 1)
    response = NPSSurveyResponse(user_comments=short_comment)
    assert response.user_comments == short_comment



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

import shutil
import os
import pathlib

import pytest
from unittest.mock import Mock

from proton.vpn.connection import VPNServer, ProtocolPorts
from proton.vpn.connection.vpnconfiguration import (VPNConfiguration)
from proton.vpn.backend.networkmanager.protocol.openvpn.openvpn import OVPNConfig
from proton.vpn.connection.interfaces import (Settings, VPNCredentials,
                                              VPNPubkeyCredentials, VPNServer,
                                              VPNUserPassCredentials, Features)
from proton.vpn.connection import VPNServer, ProtocolPorts

from boilerplate import (MockVpnCredentials, MockSettings, vpn_server)

CWD = str(pathlib.Path(__file__).parent.absolute())
VPNCONFIG_DIR = os.path.join(CWD, "vpnconfig")


def setup_module(module):
    if not os.path.isdir(VPNCONFIG_DIR):
        os.makedirs(VPNCONFIG_DIR)


def teardown_module(module):
    if os.path.isdir(VPNCONFIG_DIR):
        shutil.rmtree(VPNCONFIG_DIR)
@pytest.fixture
def modified_exec_env():
    from proton.utils.environment import ExecutionEnvironment
    m = ExecutionEnvironment().path_runtime
    ExecutionEnvironment.path_runtime = VPNCONFIG_DIR
    yield ExecutionEnvironment().path_runtime
    ExecutionEnvironment.path_runtime = m

@pytest.mark.parametrize("protocol", ["udp", "tcp"])
def test_ovpnconfig_with_settings(protocol, modified_exec_env, vpn_server):
    ovpn_cfg = OVPNConfig("", vpn_server, MockVpnCredentials(), MockSettings())
    ovpn_cfg._protocol = protocol
    output = ovpn_cfg.generate()
    assert ovpn_cfg._vpnserver.server_ip in output


@pytest.mark.parametrize("protocol", ["udp", "tcp"])
def test_ovpnconfig_with_certificate(protocol, modified_exec_env, vpn_server):
    credentials = MockVpnCredentials()
    ovpn_cfg = OVPNConfig("", vpn_server, MockVpnCredentials(), MockSettings())
    ovpn_cfg._protocol = protocol
    output = ovpn_cfg.generate()

    assert credentials.pubkey_credentials.certificate_pem in output
    assert credentials.pubkey_credentials.openvpn_private_key in output
    assert "auth-user-pass" not in output





"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from http import HTTPStatus
from unittest.mock import Mock, AsyncMock

import pytest

from proton.vpn.core.refresher.notifications_refresher import NotificationsRefresher
from proton.vpn.core.refresher.scheduler import RunAgain
from proton.vpn.session.notifications_fetcher import Notifications
from proton.session.exceptions import ProtonAPIError, ProtonAPINotReachable, ProtonAPINotAvailable


@pytest.mark.asyncio
async def test_refresh_fetches_notifications_and_returns_next_refresh_delay():
    session_holder = Mock()
    session = session_holder.session

    refresher = NotificationsRefresher(session_holder=session_holder)

    new_notifications = Mock()
    new_notifications.seconds_until_expiration = 60
    session.fetch_notifications = AsyncMock(return_value=new_notifications)

    next_refresh_delay = await refresher.refresh()

    session.fetch_notifications.assert_called_once()
    assert next_refresh_delay == RunAgain.after_seconds(new_notifications.seconds_until_expiration)


@pytest.mark.asyncio
async def test_refresh_schedules_next_refresh_when_api_returns_429():
    session_holder = AsyncMock()
    session = session_holder.session

    refresher = NotificationsRefresher(session_holder=session_holder)

    session.fetch_notifications.side_effect = ProtonAPIError(
        http_code=HTTPStatus.TOO_MANY_REQUESTS,
        http_headers={},
        json_data={"Code": HTTPStatus.TOO_MANY_REQUESTS, "Error": "Too many requests"}
    )

    try:
        await refresher.refresh()
    except ProtonAPIError:
        assert False, "ProtonAPIError was raised instead of handled"


@pytest.mark.asyncio
async def test_refresh_schedules_next_refresh_when_api_not_reachable():
    session_holder = Mock()
    session = session_holder.session

    refresher = NotificationsRefresher(session_holder=session_holder)

    session.fetch_notifications = AsyncMock(side_effect=ProtonAPINotReachable("error"))

    try:
        await refresher.refresh()
    except ProtonAPINotReachable:
        assert False, "ProtonAPINotReachable was raised instead of handled"


@pytest.mark.asyncio
async def test_refresh_schedules_next_refresh_when_api_not_available():
    session_holder = Mock()
    session = session_holder.session

    refresher = NotificationsRefresher(session_holder=session_holder)

    session.fetch_notifications = AsyncMock(side_effect=ProtonAPINotAvailable("error"))

    try:
        await refresher.refresh()
    except ProtonAPINotAvailable:
        assert False, "ProtonAPINotAvailable was raised instead of handled"


@pytest.mark.asyncio
async def test_refresh_raises_on_unexpected_error():
    session_holder = Mock()
    session = session_holder.session

    refresher = NotificationsRefresher(session_holder=session_holder)

    session.fetch_notifications = AsyncMock(side_effect=RuntimeError("unexpected"))

    with pytest.raises(RuntimeError):
        await refresher.refresh()


@pytest.mark.asyncio
async def test_update_if_necessary_skips_fetch_when_not_expired():
    session_holder = Mock()
    session = session_holder.session
    session.notifications.is_expired = False
    session.fetch_notifications = AsyncMock()

    refresher = NotificationsRefresher(session_holder=session_holder)
    await refresher.update_if_necessary()

    session.fetch_notifications.assert_not_called()


@pytest.mark.asyncio
async def test_update_if_necessary_fetches_when_expired():
    session_holder = Mock()
    session = session_holder.session
    session.notifications.is_expired = True
    session.fetch_notifications = AsyncMock()

    refresher = NotificationsRefresher(session_holder=session_holder)
    await refresher.update_if_necessary()

    session.fetch_notifications.assert_called_once()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock, AsyncMock

from proton.vpn.connection.publisher import Publisher
import pytest


@pytest.fixture
def subscriber():
    return Mock()


def test_register_registers_subscriber_if_it_was_not_registered_yet(subscriber):
    publisher = Publisher()
    publisher.register(subscriber)
    assert publisher.is_subscriber_registered(subscriber)


def test_register_does_nothing_if_the_subscriber_was_already_registered():
    subscriber = Mock()
    publisher = Publisher(subscribers=[subscriber])
    publisher.register(subscriber)
    assert publisher.number_of_subscribers == 1


def test_register_raises_value_error_if_subscriber_is_not_callable():
    publisher = Publisher()
    with pytest.raises(ValueError):
        publisher.register(None)


def test_unregister_unregisters_subscriber_if_it_was_already_registered(subscriber):
    publisher = Publisher(subscribers=[subscriber])
    publisher.unregister(subscriber)
    assert not publisher.is_subscriber_registered(subscriber)


def test_unregister_does_nothing_if_subscriber_was_never_registered():
    publisher = Publisher()
    publisher.unregister(Mock())
    assert publisher.number_of_subscribers == 0


@pytest.mark.asyncio
async def test_notify_notifies_all_registered_subscribers():
    subscribers = [Mock(), AsyncMock()]
    publisher = Publisher(subscribers=subscribers)
    publisher.notify("arg1", arg2="arg2")
    for subscriber in subscribers:
        subscriber.assert_called_with("arg1", arg2="arg2")


@pytest.mark.asyncio
async def test_notify_catches_and_logs_exceptions_when_notifying_subscribers(caplog):
    subscribers = [Mock(side_effect=RuntimeError("Bad stuff")), Mock()]
    publisher = Publisher(subscribers=subscribers)

    publisher.notify("foo")

    # Assert that, even though the first subscriber raised a RuntimeError,
    # the second one was also notified.
    for subscriber in subscribers:
        subscriber.assert_called_with("foo")

    # Assert that the error was logged.
    errors = [record for record in caplog.records if record.levelname == "ERROR"]
    assert errors
    assert errors[0].msg.startswith("An error occurred notifying subscriber")


"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass
from typing import List


@dataclass
class VPNServer:
    server_ip: str = None
    openvpn_ports: object = None
    wireguard_ports: object = None
    x25519pk: str = None
    domain: str = None
    servername: str = None


@dataclass
class VPNPubKeyCredentials:
    certificate_pem: str = None
    wg_private_key: str = None
    openvpn_private_key: str = None


@dataclass
class VPNUserPassCredentials:
    username: str = None
    password: str = None


@dataclass
class VPNCredentials:
    pubkey_credentials: VPNPubKeyCredentials = None
    userpass_credentials: VPNUserPassCredentials = None


@dataclass
class Features:
    netshield: int = None
    vpn_accelerator: bool = None
    port_forwarding: bool = None
    random_nat: bool = None
    safe_mode: bool = None


@dataclass
class Settings:
    dns_custom_ips: List[str] = None
    split_tunneling_ips: List[str] = None
    ipv6: bool = None
    features: Features = None



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from getpass import getuser

import gi
gi.require_version("NM", "1.0")
from gi.repository import NM

from proton.vpn.backend.networkmanager.killswitch.default.killswitch_connection import (
    KillSwitchConnection, KillSwitchGeneralConfig, KillSwitchIPConfig
)


def _make_general_config(permanent: bool) -> KillSwitchGeneralConfig:
    return KillSwitchGeneralConfig(
        human_readable_id="pvpn-killswitch",
        interface_name="pvpnksintrf0",
        permanent=permanent,
    )


def _make_ip_config() -> KillSwitchIPConfig:
    return KillSwitchIPConfig(
        addresses=["100.85.0.1/24"],
        dns=["0.0.0.0"],
        dns_priority=-1400,
        gateway="100.85.0.1",
        ignore_auto_dns=True,
        route_metric=98,
    )


def _make_ipv6_config() -> KillSwitchIPConfig:
    return KillSwitchIPConfig(
        addresses=["fdeb:446c:912d:08da::/64"],
        dns=["::1"],
        dns_priority=-1400,
        gateway="fdeb:446c:912d:08da::1",
        ignore_auto_dns=True,
        route_metric=95,
    )


def _get_setting_connection(connection: NM.Connection) -> NM.SettingConnection:
    return connection.get_setting(NM.SettingConnection)


def test_non_permanent_kill_switch_is_scoped_to_current_user():
    """Non-permanent kill switch should add a user permission to avoid polkit auth."""
    ks = KillSwitchConnection(
        general_settings=_make_general_config(permanent=False),
        ipv4_settings=_make_ip_config(),
        ipv6_settings=_make_ipv6_config(),
    )

    s_con = _get_setting_connection(ks.connection)

    assert s_con.get_num_permissions() == 1
    assert s_con.permissions_user_allowed(getuser())


def test_permanent_kill_switch_is_a_system_connection():
    """Permanent kill switch should NOT add a user permission (system-wide connection)."""
    ks = KillSwitchConnection(
        general_settings=_make_general_config(permanent=True),
        ipv4_settings=_make_ip_config(),
        ipv6_settings=_make_ipv6_config(),
    )

    s_con = _get_setting_connection(ks.connection)

    assert s_con.get_num_permissions() == 0



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytest
from unittest.mock import Mock, AsyncMock

from proton.vpn.backend.networkmanager.core import LocalAgentMixin
from proton.vpn.core.settings.features import Features
from proton.vpn.session.servers.types import TierEnum


@pytest.fixture
def mixin_mock():
    mock_listener = Mock()
    mock_listener.request_features = AsyncMock()

    mixin_mock = LocalAgentMixin(user_tier=TierEnum.FREE)
    mixin_mock._agent_listener = mock_listener
    mixin_mock._vpnserver = Mock(label="TEST#1")
    return mixin_mock


@pytest.mark.asyncio
async def test_request_features_is_skipped_when_user_is_free(mixin_mock):
    mixin_mock._user_tier = TierEnum.FREE
    features = Features.default(user_tier=0)

    await mixin_mock._request_connection_features(features)

    mixin_mock._agent_listener.request_features.assert_not_called()


@pytest.mark.asyncio
async def test_request_features_is_called_when_user_tier_is_not_free(mixin_mock):
    mixin_mock._user_tier = TierEnum.PLUS
    features = Features.default(user_tier=0)
    await mixin_mock._request_connection_features(features)

    mixin_mock._agent_listener.request_features.assert_called_once()




"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
import os
import pytest
from types import SimpleNamespace
import tempfile
import json

from proton.vpn.core.session_holder import ClientTypeMetadata
from proton.vpn.core.usage import UsageReporting

SECRET_FILE = "secret.txt"
SECRET_PATH = os.path.join("/home/wozniak/5nkfiudfmk/.cache", SECRET_FILE)
MACHINE_ID = "bg77t2rmpjhgt9zim5gkz4t78jfur39f"
SENTRY_USER_ID = "70cf75689cecae78ec588316320d76477c71031f7fd172dd5577ac95934d4499"
USERNAME = "tester"
EMAIL = "toby.tubface@bucketworld.com"
GPG_KEY_ID = "D212342HE93E32P8"

EVENT_WITH_USERNAME = {
    "frames": [
        {
            "filename": "/home/tester/src/quick_connect_widget.py",
            "abs_path": "/home/tester/src/quick_connect_widget.py",
            "function": "_on_disconnect_button_clicked",
            "module": "proton.vpn.app.gtk.widgets.vpn.quick_connect_widget",
            "lineno": 102,
            "pre_context": [
                "        future = self._controller.connect_to_fastest_server()",
                "        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.",
                "",
                "    def _on_disconnect_button_clicked(self, _):",
                "        logger.info(\"Disconnect from VPN\", category=\"ui\", event=\"disconnect\")"
            ],
            "context_line": "        future = self._controller.disconnect()",
            "post_context": [
                "        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any."
            ],
            "vars": {
                "self": "<quick_connect_widget.... (proton+... at 0x601b5e322d90)>",
                "_": "<Gtk.Button object at 0x7f83d4aa8140 (GtkButton at 0x601b5e37ea40)>"
            },
            "in_app": True
        },
        {
            "filename": "/home/tester/src/ProtonVPN/linux/proton-vpn-gtk-app/proton/vpn/app/gtk/controller.py",
            "abs_path": "/home/tester/src/ProtonVPN/linux/proton-vpn-gtk-app/proton/vpn/app/gtk/controller.py",
            "function": "disconnect",
            "module": "proton.vpn.app.gtk.controller",
            "lineno": 224,
            "pre_context": [
                "        :return: A Future object that resolves once the connection reaches the",
                "        \"disconnected\" state.",
                "        \"\"\"",
                "        error = FileNotFoundError(\"This method is not implemented\")",
                "        error.filename = \"/home/wozniak/randomfile.py\""
            ],
            "context_line": "        raise error",
            "post_context": [
                "",
                "        return self.executor.submit(self._connector.disconnect)",
                "",
                "    @property",
                "    def account_name(self) -> str:"
            ],
            "vars": {
                "self": "<proton.vpn.app.gtk.controller.Controller object at 0x7f83e1856da0>",
                "error": "FileNotFoundError('This method is not implemented')"
            },
            "in_app": True
        }
    ]
}

EVENT_WITH_GPG_ID = {
    "exception": {
        "values": [
            {
                "type": "DBusErrorResponse",
                "value": "[me.grimsteel.PassSecretService.GPGError] ('gpg: encrypted with cv25519 key, ID D212342HE93E32P8, created 2025-02-03\\n      \"Toby (pass & proton) <toby.tubface@bucketworld.com>\"\\ngpg: [don\\'t know]: invalid packet (ctb=6c)\\ngpg: packet(5) with unknown version 18\\n',)",
            }
        ]
    }
}

EVENT_WITH_EMAIL = {
    "exception": {
        "values": [
            {
                "type": "DBusErrorResponse",
                "value": "Some error that includes an email address: toby.tubface@bucketworld.com',)",
            }
        ]
    }
}

@pytest.mark.parametrize("enabled", [True, False])
def test_usage_report_enabled(enabled):
    report_error = SimpleNamespace(invoked=False)

    usage_reporting = UsageReporting(ClientTypeMetadata("test_usage.py", "none"))

    def capture_exception(error):
        report_error.invoked = True

    usage_reporting.enabled = enabled
    usage_reporting._capture_exception = capture_exception

    EMPTY_ERROR = None
    usage_reporting.report_error(EMPTY_ERROR)

    assert report_error.invoked == enabled, "UsageReporting enable state does not match the error reporting"


def test_userid_calaculation():
    with tempfile.NamedTemporaryFile() as file:
        file.write(MACHINE_ID.encode('utf-8'))
        file.seek(0)

        assert UsageReporting._get_user_id(
            machine_id_filepath=file.name,
            user_name=USERNAME) == SENTRY_USER_ID, "Error hashing does not match the expected value"


def test_sanitize_event_for_username():

    event = copy.deepcopy(EVENT_WITH_USERNAME)

    UsageReporting._sanitize_event(event, None, USERNAME)

    assert USERNAME in json.dumps(EVENT_WITH_USERNAME), "Username should be in the event"
    assert USERNAME not in json.dumps(event), "Username should not be in the event"


def test_sanitize_event_for_email():

    event = copy.deepcopy(EVENT_WITH_EMAIL)

    UsageReporting._sanitize_event(event, None, USERNAME)

    assert EMAIL in json.dumps(EVENT_WITH_EMAIL), "Email should be in the event"
    assert EMAIL not in json.dumps(event), "Email should not be in the event"


def test_sanitize_event_for_gpg_id():

    event = copy.deepcopy(EVENT_WITH_GPG_ID)

    UsageReporting._sanitize_event(event, None, USERNAME)

    assert GPG_KEY_ID in json.dumps(EVENT_WITH_GPG_ID), "GPG ID should be in the event"
    assert GPG_KEY_ID not in json.dumps(event), "GPG ID should not be in the event"





"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from http import HTTPStatus
from unittest.mock import Mock, AsyncMock

import pytest

from proton.vpn.core.refresher.feature_flags_refresher import FeatureFlagsRefresher
from proton.vpn.core.refresher.scheduler import RunAgain
from proton.session.exceptions import ProtonAPIError


@pytest.mark.asyncio
async def test_refresh_fetches_feature_flags_and_returns_next_refresh_delay():
    session_holder = Mock()
    session = session_holder.session

    refresher = FeatureFlagsRefresher(session_holder=session_holder)

    new_feature_flags = Mock()
    new_feature_flags.seconds_until_expiration = 60
    session.fetch_feature_flags = AsyncMock()
    session.fetch_feature_flags.return_value = new_feature_flags

    next_refresh_delay = await refresher.refresh()

    session.fetch_feature_flags.assert_called_once()

    assert next_refresh_delay == RunAgain.after_seconds(new_feature_flags.seconds_until_expiration)


@pytest.mark.asyncio
async def test_refresh_schedules_next_refresh_if_feature_flags_is_expired_and_api_error_exception_is_raised():
    session_holder = Mock()
    session = session_holder.session

    refresher = FeatureFlagsRefresher(session_holder=session_holder)

    new_feature_flags = Mock()
    new_feature_flags.seconds_until_expiration = 60
    session.fetch_feature_flags = AsyncMock(side_effect=ProtonAPIError(
        http_code=HTTPStatus.TOO_MANY_REQUESTS,
        http_headers={},
        json_data={"Code": HTTPStatus.TOO_MANY_REQUESTS, "Error": "Error message"}
    ))

    next_refresh_delay = await refresher.refresh()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn.core.refresher import VPNDataRefresher
from proton.vpn.core.registry import Registry
from proton.vpn.session.servers import LogicalServer
from proton.vpn.session.client_config import ClientConfig
from proton.vpn.core.vpnconnector import VPNConnector
from proton.vpn.connection import events, exceptions, states
from unittest.mock import Mock, AsyncMock
import pytest

from tests.connection.test_vpnconnection import DummyVPNConnection


class _ProtonProtocol(DummyVPNConnection):
    protocol = "proton_protocol"

    @classmethod
    def get_protocol_group(cls):
        return "proton"


class _GenericProtocol(DummyVPNConnection):
    protocol = "generic_protocol"

    @classmethod
    def get_protocol_group(cls):
        return "generic"


LOGICAL_SERVER_DATA = {
  "Name": "IS#1",
  "ID": "OYB-3pMQQA2Z2Qnp5s5nIvTVO2alU6h82EGLXYHn1mpbsRvE7UfyAHbt0_EilRjxhx9DCAUM9uXfM2ZUFjzPXw==",
  "Status": 1,
  "Servers": [
    {
      "EntryIP": "185.159.158.1",
      "Domain": "node-is-01.protonvpn.net",
      "X25519PublicKey": "yKbYe2XwbeNN9CuPZcwMF/lJp6a62NEGiHCCfpfxrnE=",
      "Status": 1,
    }
  ],
  "Label": "3",
}


def test_get_vpn_server_returns_vpn_server_built_from_logical_server_and_client_config():
    vpn_connector_wrapper = VPNConnector(
      session_holder=None,
      settings_persistence=None,
      usage_reporting=None,
      registry=Registry()
    )

    logical_server = LogicalServer(data=LOGICAL_SERVER_DATA)
    client_config = ClientConfig.default()

    vpn_server = vpn_connector_wrapper.get_vpn_server(logical_server, client_config)

    physical_server = logical_server.physical_servers[0]
    assert vpn_server.server_ip == physical_server.entry_ip
    assert vpn_server.domain == physical_server.domain
    assert vpn_server.x25519pk == physical_server.x25519_pk
    assert vpn_server.openvpn_ports.udp == client_config.openvpn_ports.udp
    assert vpn_server.openvpn_ports.tcp == client_config.openvpn_ports.tcp
    assert vpn_server.wireguard_ports.udp == client_config.wireguard_ports.udp
    assert vpn_server.wireguard_ports.tcp == client_config.wireguard_ports.tcp
    assert vpn_server.server_id == logical_server.id
    assert vpn_server.server_name == logical_server.name
    assert vpn_server.label == physical_server.label


@pytest.mark.asyncio
async def test__on_connection_event_swallows_and_does_not_report_policy_errors():
    vpn_connector_wrapper = VPNConnector(
      session_holder=None,
      settings_persistence=None,
      usage_reporting=Mock(),
      registry=Registry(),
      state=states.Connected(),
    )

    event = events.Disconnected()
    event.context.error = exceptions.FeaturePolicyError("Policy error")

    await vpn_connector_wrapper._on_connection_event(event)

    vpn_connector_wrapper._usage_reporting.report_error.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize("error", [
    exceptions.FeatureError("generic feature error"),
    exceptions.FeatureSyntaxError("Feature syntax error")
])
async def test__on_connection_event_reports_feature_syntax_errors_but_no_other_feature_error(error):
    vpn_connector_wrapper = VPNConnector(
        session_holder=None,
        settings_persistence=None,
        usage_reporting=Mock(),
        registry=Registry(),
        state=states.Connected(),
    )

    event = events.Disconnected()
    event.context.error = error

    await vpn_connector_wrapper._on_connection_event(event)
    if isinstance(error, exceptions.FeatureSyntaxError):
        vpn_connector_wrapper._usage_reporting.report_error.assert_called_once_with(event.context.error)
    elif isinstance(error, exceptions.FeatureError):
        vpn_connector_wrapper._usage_reporting.report_error.assert_not_called()
    else:
        raise ValueError(f"Unexpected test parameter: {error}")

@pytest.mark.asyncio
async def test__on_connection_event_reports_unexpected_exceptions_and_bubbles_them_up():
    vpn_connector_wrapper = VPNConnector(
        session_holder=None,
        settings_persistence=None,
        usage_reporting=Mock(),
        registry=Registry(),
        state=states.Connected(),
    )

    event = events.Disconnected()
    event.context.error = Exception("Unexpected error")

    with pytest.raises(Exception):
        await vpn_connector_wrapper._on_connection_event(event)

    vpn_connector_wrapper._usage_reporting.report_error.assert_called_once_with(event.context.error)


def test_on_state_change_stores_new_device_ip_when_successfully_connected_to_vpn_and_connection_details_and_device_ip_are_set():
    publisher_mock = Mock()
    session_holder_mock = Mock()
    new_connection_details = events.ConnectionDetails(
        device_ip="192.168.0.1",
        device_country="PT",
        server_ipv4="0.0.0.0",
        server_ipv6=None,
    )
    _ = VPNConnector(
        session_holder=session_holder_mock,
        settings_persistence=None,
        usage_reporting=None,
        registry=Registry(),
        connection_persistence=Mock(),
        publisher=publisher_mock,
        port_forward_file_handler=Mock()
    )

    on_state_change_update_location_callback = publisher_mock.register.call_args_list[0][0][0]

    connected_event = events.Connected(
        context=events.EventContext(
            connection=Mock(),
            connection_details=new_connection_details
        )
    )
    connected_state = states.Connected(context=states.StateContext(connected_event))

    on_state_change_update_location_callback(connected_state)

    vpn_location = session_holder_mock.session.set_location.call_args[0][0]

    session_holder_mock.session.set_location.assert_called_once()
    assert vpn_location.IP == new_connection_details.device_ip


def test_on_state_change_notifies_port_forwarding_file_handler_with_new_state():
    publisher_mock = Mock()
    session_holder_mock = Mock()
    port_forwarding_file_handler_mock = Mock(name="port_forwarding_file_handler_mock")
    new_connection_details = events.ConnectionDetails(
        device_ip="192.168.0.1",
        device_country="PT",
        server_ipv4="0.0.0.0",
        server_ipv6=None,
    )
    _ = VPNConnector(
        session_holder=session_holder_mock,
        settings_persistence=None,
        usage_reporting=None,
        registry=Registry(),
        connection_persistence=Mock(),
        publisher=publisher_mock,
        port_forward_file_handler=port_forwarding_file_handler_mock
    )

    on_state_change_update_port_callback = publisher_mock.register.call_args_list[1][0][0]

    connected_event = events.Connected(
        context=events.EventContext(
            connection=Mock(),
            connection_details=new_connection_details
        )
    )
    connected_state = states.Connected(context=states.StateContext(connected_event))
    on_state_change_update_port_callback(connected_state)

    port_forwarding_file_handler_mock.on_state_change_update_port.assert_called_once_with(connected_state)


def test_on_state_change_skip_store_new_device_ip_when_successfully_connected_to_vpn_and_connection_details_is_none():
    publisher_mock = Mock()
    session_holder_mock = Mock()
    _ = VPNConnector(
        session_holder=session_holder_mock,
        settings_persistence=None,
        usage_reporting=None,
        registry=Registry(),
        connection_persistence=Mock(),
        publisher=publisher_mock
    )

    on_state_change_callback = publisher_mock.register.call_args[0][0]

    connected_event = events.Connected(
        context=events.EventContext(
            connection=Mock(),
            connection_details=None
        )
    )
    connected_state = states.Connected(context=states.StateContext(connected_event))

    on_state_change_callback(connected_state)

    session_holder_mock.session.set_location.assert_not_called()


def test_on_state_change_skip_store_new_device_ip_when_successfully_connected_to_vpn_and_device_ip_is_none():
    publisher_mock = Mock()
    session_holder_mock = Mock()
    new_connection_details = events.ConnectionDetails(
        device_ip=None,
        device_country="PT",
        server_ipv4="0.0.0.0",
        server_ipv6=None,
    )
    _ = VPNConnector(
        session_holder=session_holder_mock,
        settings_persistence=None,
        usage_reporting=None,
        registry=Registry(),
        connection_persistence=Mock(),
        publisher=publisher_mock
    )

    on_state_change_callback = publisher_mock.register.call_args[0][0]

    connected_event = events.Connected(
        context=events.EventContext(
            connection=Mock(),
            connection_details=new_connection_details
        )
    )
    connected_state = states.Connected(context=states.StateContext(connected_event))

    on_state_change_callback(connected_state)

    session_holder_mock.session.set_location.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize("state_class, update_credentials_expected", [
    (states.Connected, True),
    (states.Error, True),
    (states.Disconnected, False),
    (states.Connecting, False),
    (states.Disconnecting, False),
])
async def test_connector_updates_connection_credentials_when_certificate_is_refreshed_and_current_state_is_connected_or_error(
        state_class, update_credentials_expected
):
    session_holder = Mock()
    current_state = state_class(states.StateContext(connection=AsyncMock()))

    connector = VPNConnector(
        session_holder=session_holder,
        settings_persistence=Mock(),
        usage_reporting=Mock(),
        registry=Registry(),
        connection_persistence=Mock(),
        state=current_state
    )

    refresher = VPNDataRefresher(session_holder=session_holder, scheduler=Mock())
    connector.subscribe_to_certificate_updates(refresher)

    # Trigger certificated updated callback
    await refresher._certificate_refresher.certificate_updated_callback()

    assert current_state.context.connection.update_credentials.called is update_credentials_expected

    if update_credentials_expected:
        current_state.context.connection.update_credentials.assert_called_once_with(session_holder.vpn_credentials)


def test_iter_available_protocols_returns_only_protocols_matching_group():
    registry = Registry()
    registry.register(_ProtonProtocol)
    registry.register(_GenericProtocol)

    connector = VPNConnector(
        session_holder=None,
        settings_persistence=None,
        usage_reporting=None,
        registry=registry,
    )

    result = list(connector.iter_available_protocols("proton"))

    assert result == [_ProtonProtocol]


def test_iter_available_protocols_returns_empty_when_no_protocols_match_group():
    registry = Registry()
    registry.register(_ProtonProtocol)
    registry.register(_GenericProtocol)

    connector = VPNConnector(
        session_holder=None,
        settings_persistence=None,
        usage_reporting=None,
        registry=registry,
    )

    result = list(connector.iter_available_protocols("nonexistent"))

    assert result == []



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

import pytest
from unittest.mock import Mock

from proton.vpn.connection import VPNServer, ProtocolPorts
from proton.vpn.connection.interfaces import (Settings, VPNCredentials,
                                              VPNPubkeyCredentials, VPNServer,
                                              VPNUserPassCredentials)


class MockVPNPubkeyCredentials(VPNPubkeyCredentials):
    @property
    def certificate_pem(self):
        return "pem-cert"

    @property
    def wg_private_key(self):
        return "wg-private-key"

    @property
    def openvpn_private_key(self):
        return "ovpn-private-key"

    def get_ed25519_sk_pem(self, password=None):
        return "encrypted-ovpn-private-key"


class MockVPNUserPassCredentials(VPNUserPassCredentials):
    @property
    def username(self):
        return "test-username"

    @property
    def password(self):
        return "test-password"


class MockVpnCredentials(VPNCredentials):
    @property
    def pubkey_credentials(self):
        return MockVPNPubkeyCredentials()

    @property
    def userpass_credentials(self):
        return MockVPNUserPassCredentials()


class MockSettings(Settings):
    @property
    def dns_custom_ips(self):
        return ["1.1.1.1", "10.10.10.10"]

    @property
    def features(self):
        return Mock()


@pytest.fixture
def vpn_server():
    return VPNServer(
        server_ip="10.10.1.1",
        domain="com.test-domain.www",
        x25519pk="wg_public_key",
        openvpn_ports=ProtocolPorts(
            tcp=[80, 1194], udp=[445, 5995], tls=[8080]),
        wireguard_ports=ProtocolPorts(
            tcp=[443, 88], udp=[445], tls=[8080]),
        server_name="TestServer#10",
        server_id="OYB-3pMQQA2Z2Qnp5s5nIvTVO2...lRjxhx9DCAUM9uXfM2ZUFjzPXw==",
        has_ipv6_support=False,
        label="0"

    )



#!/usr/bin/env python

from setuptools import setup, find_namespace_packages
import re

VERSIONS = 'versions.yml'
VERSION = re.search(r'version: (\S+)', open(VERSIONS, encoding='utf-8').readline()).group(1)

setup(
    name="proton-vpn-api-core",
    version=VERSION,
    description="Proton AG VPN Core API",
    author="Proton AG",
    author_email="opensource@proton.me",
    url="https://github.com/ProtonVPN/python-proton-vpn-api-core",
    include_package_data=True,
    install_requires=[
        "proton-core", "distro", "sentry-sdk",
        "cryptography", "PyNaCl", "distro", "fido2", "packaging",
        "pygobject", "pycairo", "jinja2", "proton-vpn-local-agent"  # network manager backend
    ],
    extras_require={
        "development": ["pytest", "pytest-coverage", "pylint", "flake8", "pytest-asyncio", "PyYAML"]
    },
    packages=find_namespace_packages(include=[
        "proton.vpn.core*",
        "proton.vpn.connection*",
        "proton.vpn.killswitch.interface*",
        "proton.vpn.session*",
        "proton.vpn.logging*",
        "proton.vpn.split_tunneling*",
        "proton.vpn.backend.networkmanager.core*",
        "proton.vpn.backend.networkmanager.protocol.openvpn*",
        "proton.vpn.backend.networkmanager.protocol.wireguard*",
        "proton.vpn.backend.networkmanager.protocol.protun*",
        "proton.vpn.backend.networkmanager.killswitch.default*",
        "proton.vpn.backend.networkmanager.killswitch.wireguard*",
    ]),
    entry_points={
        "proton_loader_killswitch": [
            "default = proton.vpn.backend.networkmanager.killswitch.default:NMKillSwitch",
            "wireguard = proton.vpn.backend.networkmanager.killswitch.wireguard:WGKillSwitch",
        ]
    },
    python_requires=">=3.9",
    license="GPLv3",
    platforms="Linux",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Security",
    ]
)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import os

import pytest

from proton.vpn.connection import VPNServer, ProtocolPorts
from proton.vpn.connection.vpnconfiguration import VPNConfiguration

from .common import (CWD, MockSettings, MockVpnCredentials)
import shutil

VPNCONFIG_DIR = os.path.join(CWD, "vpnconfig")


def setup_module(module):
    if not os.path.isdir(VPNCONFIG_DIR):
        os.makedirs(VPNCONFIG_DIR)


def teardown_module(module):
    if os.path.isdir(VPNCONFIG_DIR):
        shutil.rmtree(VPNCONFIG_DIR)


@pytest.fixture
def modified_exec_env():
    from proton.utils.environment import ExecutionEnvironment
    m = ExecutionEnvironment().path_runtime
    ExecutionEnvironment.path_runtime = VPNCONFIG_DIR
    yield ExecutionEnvironment().path_runtime
    ExecutionEnvironment.path_runtime = m


@pytest.fixture
def vpn_server():
    return VPNServer(
        server_ip="10.10.1.1",
        domain="com.test-domain.www",
        x25519pk="wg_public_key",
        openvpn_ports=ProtocolPorts(
            tcp=[80, 1194], udp=[445, 5995], tls=[8080]),
        wireguard_ports=ProtocolPorts(
            tcp=[443, 88], udp=[445], tls=[8080]),
        server_name="TestServer#10",
        server_id="OYB-3pMQQA2Z2Qnp5s5nIvTVO2...lRjxhx9DCAUM9uXfM2ZUFjzPXw==",
        has_ipv6_support=False,
        label="0"

    )


class MockVpnConfiguration(VPNConfiguration):
    extension = ".test-extension"

    def generate(self):
        return "test-content"


def test_not_implemented_generate(vpn_server):
    cfg = VPNConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    with pytest.raises(NotImplementedError):
        cfg.generate()


def test_ensure_configuration_file_is_created(modified_exec_env, vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    with cfg as f:
        assert os.path.isfile(f)


def test_ensure_configuration_file_is_deleted(vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    fp = None
    with cfg as f:
        fp = f
        assert os.path.isfile(fp)

    assert not os.path.isfile(fp)


def test_ensure_generate_is_returning_expected_content(vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    with cfg as f:
        with open(f) as _f:
            line = _f.readline()
            _cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
            assert line == _cfg.generate()


def test_ensure_same_configuration_file_in_case_of_duplicate(vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    with cfg as f:
        with cfg as _f:
            assert os.path.isfile(f) and os.path.isfile(_f) and f == _f


@pytest.mark.parametrize(
    "expected_mask, cidr", [
        ("0.0.0.0", "0"),
        ("255.0.0.0", "8"),
        ("255.255.0.0", "16"),
        ("255.255.255.0", "24"),
        ("255.255.255.255", "32")
    ]
)
def test_cidr_to_netmask(cidr, expected_mask, vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    assert cfg.cidr_to_netmask(cidr) == expected_mask


@pytest.mark.parametrize("ipv4", ["192.168.1.1", "109.162.10.9", "1.1.1.1", "10.10.10.10"])
def test_valid_ips(ipv4, vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    cfg.is_valid_ipv4(ipv4)


@pytest.mark.parametrize("ipv4", ["192.168.1.90451", "109.", "1.-.1.1", "1111.10.10.10"])
def test_not_valid_ips(ipv4, vpn_server):
    cfg = MockVpnConfiguration(vpn_server, MockVpnCredentials(), MockSettings())
    cfg.is_valid_ipv4(ipv4)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from http import HTTPStatus
from unittest.mock import Mock, AsyncMock, patch

import pytest

from proton.vpn.core.refresher.scheduler import RunAgain
from proton.vpn.core.refresher.server_list_refresher import ServerListRefresher

from proton.session.exceptions import ProtonAPIError



@pytest.mark.asyncio
async def test_refresh_fetches_server_list_if_expired_and_returns_next_refresh_delay():
    session_holder = Mock()
    session = session_holder.session

    # The current server list is expired.
    session.server_list.expired = True

    new_server_list = Mock()
    new_server_list.seconds_until_expiration = 15 * 60
    session.fetch_server_list = AsyncMock()
    session.fetch_server_list.return_value = new_server_list

    refresher = ServerListRefresher(session_holder=session_holder)
    refresher.server_list_updated_callback = Mock()

    next_refresh_delay = await refresher.refresh()

    # A new server list should've been fetched.
    session.fetch_server_list.assert_called_once()

    # The callback to notify of server list updates should have been called.
    refresher.server_list_updated_callback.assert_called_once_with()

    # And the new refresh should've been scheduled after the new
    # server list/loads expire again.
    assert next_refresh_delay == RunAgain.after_seconds(new_server_list.seconds_until_expiration)


@pytest.mark.asyncio
async def test_refresh_updates_server_loads_if_expired_and_returns_next_refresh_delay():
    session_holder = Mock()
    session = session_holder.session

    # Only loads are expired
    session.server_list.expired = False
    session.server_list.loads_expired = True

    updated_server_list = Mock()
    updated_server_list.seconds_until_expiration = 60
    session.update_server_loads = AsyncMock()
    session.update_server_loads.return_value = updated_server_list

    refresher = ServerListRefresher(session_holder=session_holder)
    refresher.server_loads_updated_callback = Mock()

    next_refresh_delay = await refresher.refresh()

    # The server list should not have been fetched...
    session.fetch_server_list.assert_not_called()
    # but the loads should have been updated.
    session.update_server_loads.assert_called_once()

    # The callback to notify of server load updates should have been called.
    refresher.server_loads_updated_callback.assert_called_once_with()

    # And the next refresh should've been scheduled when the updated
    # server list expires.
    assert next_refresh_delay == RunAgain.after_seconds(updated_server_list.seconds_until_expiration)


@pytest.mark.asyncio
async def test_refresh_schedules_next_refresh_if_server_list_is_not_expired():
    session_holder = Mock()
    session = session_holder.session

    # The current server list is not expired.
    session.server_list.expired = False
    session.server_list.loads_expired = False
    session.server_list.seconds_until_expiration = 60

    refresher = ServerListRefresher(session_holder=session_holder)

    next_refresh_delay = await refresher.refresh()

    # The server list should not have been fetched.
    session.fetch_server_list.assert_not_called()
    # The server loads should not have been fetched either.
    session.update_server_loads.assert_not_called()

    # And the next refresh should've been scheduled when the current
    # server list expires.
    assert next_refresh_delay == RunAgain.after_seconds(session.server_list.seconds_until_expiration)


@pytest.mark.asyncio
@patch("proton.vpn.core.refresher.server_list_refresher.ServerList.get_loads_refresh_interval_in_seconds")
async def test_refresh_schedules_next_refresh_if_server_list_is_expired_and_api_error_exception_is_raised(mock_get_loads_refresh_interval_in_seconds):
    get_loads_refresh_interval_in_seconds = 10
    mock_get_loads_refresh_interval_in_seconds.return_value = get_loads_refresh_interval_in_seconds
    session_holder = Mock()
    session = session_holder.session

    # The current server list is expired.
    session.server_list.expired = True

    # Mock a 429 response from the server
    session.fetch_server_list = AsyncMock(side_effect=ProtonAPIError(
        http_code=HTTPStatus.TOO_MANY_REQUESTS,
        http_headers={},
        json_data={"Code": HTTPStatus.TOO_MANY_REQUESTS, "Error": "Error message"}
    ))

    refresher = ServerListRefresher(session_holder=session_holder)
    refresher.server_list_updated_callback = Mock()

    # No error message should be raised since it's been swallowed
    next_refresh_delay = await refresher.refresh()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from unittest.mock import Mock, AsyncMock, call
import pytest
from ipaddress import ip_network, collapse_addresses

from proton.vpn.backend.networkmanager.killswitch.default import NMKillSwitch
from proton.vpn.backend.networkmanager.killswitch.default.killswitch_connection_handler \
    import build_routes_list, LOCAL_AGENT_SERVER_ADDR


@pytest.fixture
def vpn_server():
    vpn_server_mock = Mock()
    vpn_server_mock.server_ip = "1.1.1.1"

    return vpn_server_mock


@pytest.mark.asyncio
async def test_enable_without_vpn_server_adds_full_ks_and_removes_routed_ks():
    ks_handler_mock = AsyncMock()
    nm_killswitch = NMKillSwitch(ks_handler_mock)

    await nm_killswitch.enable()

    assert ks_handler_mock.method_calls == [
        call.add_full_killswitch_connection(False),
        call.remove_routed_killswitch_connection(),
    ]


@pytest.mark.asyncio
async def test_enable_with_vpn_server(vpn_server):
    """
    When enabling the KS specifying a vpn server to connect to we expect:
     1) The full KS is added first, to block all network traffic until the routed KS is set up.
     2) The routed KS is removed (if found).
     2) A new routed KS whitelisting the VPN server IP is added.
     4) The full KS is removed to let the routed KS take over.
    """
    ks_handler_mock = AsyncMock()
    nm_killswitch = NMKillSwitch(ks_handler_mock)

    await nm_killswitch.enable(vpn_server)

    assert ks_handler_mock.method_calls == [
        call.add_full_killswitch_connection(False),
        call.remove_routed_killswitch_connection(),
        call.add_routed_killswitch_connection(vpn_server.server_ip, False),
        call.remove_full_killswitch_connection()
    ]


@pytest.mark.asyncio
async def test_disable_killswitch_removes_full_and_routed_ks():
    ks_handler_mock = AsyncMock()
    nm_killswitch = NMKillSwitch(ks_handler_mock)

    await nm_killswitch.disable()

    assert ks_handler_mock.method_calls == [
        call.remove_full_killswitch_connection(),
        call.remove_routed_killswitch_connection()
    ]


@pytest.mark.asyncio
async def test_enable_ipv6_leak_protection_adds_ipv6_ks():
    ks_handler_mock = AsyncMock()

    nm_killswitch = NMKillSwitch(ks_handler_mock)
    await nm_killswitch.enable_ipv6_leak_protection()

    assert ks_handler_mock.method_calls == [
        call.add_ipv6_leak_protection()
    ]


@pytest.mark.asyncio
async def test_disable_ipv6_leak_protection_removes_ipv6_ks():
    ks_handler_mock = AsyncMock()

    nm_killswitch = NMKillSwitch(ks_handler_mock)
    await nm_killswitch.disable_ipv6_leak_protection()

    assert ks_handler_mock.method_calls == [
        call.remove_ipv6_leak_protection()
    ]


def test_build_routes_list():
    vpn_server = "192.168.2.1"

    allowed_routes = [
        ip_network(LOCAL_AGENT_SERVER_ADDR),
        ip_network(vpn_server)
    ]
    forbidden_routes = list(build_routes_list(vpn_server))
    all_routes = list(collapse_addresses(allowed_routes + forbidden_routes))

    assert all_routes == [ip_network('0.0.0.0/0')]



import pytest

from proton.vpn.session.utils import to_semver_build_metadata_format, semver_from_pep440


@pytest.mark.parametrize("input,expected_output", [
    ("x86_64", "x86-64"),  # Underscores are replaced by hyphens
    ("aarch64", "aarch64"),
    ("!@#$%^&*()+=<>~,./?\\|[]{} ", ""),  # Only alphanumeric characters and hyphens allowed.
    ("", ""),
    (None, None)
])
def test_to_semver_build_metadata_format(input, expected_output):
    assert to_semver_build_metadata_format(input) == expected_output


@pytest.mark.parametrize("pep440_version, expected_semver_version", [
    ("1.2.3", "1.2.3"),
    ("1.2.3a4", "1.2.3-alpha.4"),
    ("1.2.3b4", "1.2.3-beta.4"),
    ("1.2.3rc4", "1.2.3-rc.4"),
    ("1.2.3a4.dev5+abc", "1.2.3-alpha.4-dev.5+abc")
])
def test_from_pep440(pep440_version, expected_semver_version):
    result = semver_from_pep440(pep440_version)
    assert result == expected_semver_version



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

import proton.vpn.session.fetcher as fetcher
from proton.vpn.session.fetcher import VPNSessionFetcher
from proton.vpn.core.settings import SplitTunnelingConfig
from proton.vpn.core.settings.features import Features


def test_extract_features():

    actual = VPNSessionFetcher._convert_features(
        Features(
            netshield=2,
            moderate_nat=False,
            vpn_accelerator=False,
            port_forwarding=True,
            split_tunneling=SplitTunnelingConfig()
        )
    )

    expected = {
        fetcher.API_NETSHIELD: 2,
        fetcher.API_VPN_ACCELERATOR: False,
        fetcher.API_MODERATE_NAT: False,
        fetcher.API_PORT_FORWARDING: True,
    }

    assert actual == expected



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import IntEnum
from typing import List, TYPE_CHECKING
from pathlib import Path

from proton.utils.environment import VPNExecutionEnvironment
from proton.vpn.session.exceptions import NotificationDecodeError
from proton.vpn.session.utils import RefreshCalculator, rest_api_request
from proton.vpn import logging
from proton.vpn.core.cache_handler import CacheHandler

if TYPE_CHECKING:
    from proton.vpn.session.api import VPNSession

logger = logging.getLogger(__name__)

REFRESH_INTERVAL = 24 * 60 * 60  # 1 day

# pylint: disable=duplicate-code


class NotificationType(IntEnum):
    """Types of pull notification"""
    INVALID = -1
    NPSSURVEY = 4


@dataclass
class NpsSurvey:
    """Represents an NPS survey notification"""
    survey_id: str
    start_time: datetime
    end_time: datetime
    seen: bool

    @property
    def is_active(self) -> bool:
        """Returns whether this NpsSurvey is currently active"""
        now = datetime.now(tz=timezone.utc)
        return self.start_time <= now < self.end_time

    @staticmethod
    def from_dict(data: dict) -> "NpsSurvey":
        """Converts to NpsSurvey object"""
        try:
            return NpsSurvey(
                survey_id=data["NotificationID"],
                start_time=datetime.fromtimestamp(
                    data["StartTime"],
                    tz=timezone.utc
                ),
                end_time=datetime.fromtimestamp(
                    data["EndTime"],
                    tz=timezone.utc
                ),
                seen=data.get("seen", False)
            )
        except (KeyError, TypeError, ValueError) as error:
            raise NotificationDecodeError(
                "Error parsing NpsSurvey pull notification."
            ) from error


class Notifications:  # pylint: disable=too-few-public-methods
    """Contains a record of pulled notifications."""
    def __init__(self, api_data: dict):
        self._api_data = api_data
        self._expiration_time = api_data.get(
            "ExpirationTime",
            RefreshCalculator.get_expiration_time(
                refresh_interval=REFRESH_INTERVAL
            )
        )

    def get_nps_survey_notifications(self) -> List[NpsSurvey]:
        """Get a list of NPS Survey notifications."""
        nps_notifications = []
        for notification in self._api_data.get("Notifications", []):
            notification_type = notification.get("Type", NotificationType.INVALID)
            if notification_type == NotificationType.NPSSURVEY:
                try:
                    nps_notifications.append(NpsSurvey.from_dict(notification))
                except NotificationDecodeError:
                    logger.warning(
                        "NPSSurvey notification could not be deserialized.",
                        exc_info=True
                    )

        return nps_notifications

    @property
    def data(self) -> dict:
        """Returns dict with notification data"""
        return self._api_data

    @property
    def is_expired(self) -> bool:
        """Returns if data has expired"""
        return RefreshCalculator.get_is_expired(self._expiration_time)

    @property
    def seconds_until_expiration(self) -> int:
        """Returns amount of seconds until it expires."""
        return RefreshCalculator.get_seconds_until_expiration(self._expiration_time)

    @staticmethod
    def get_refresh_interval_in_seconds() -> int:
        """Returns refresh interval in seconds."""
        return RefreshCalculator(REFRESH_INTERVAL).get_refresh_interval_in_seconds()


class NotificationsFetcher:
    """Fetches and caches notifications from Proton's REST API."""
    ROUTE = "/core/v4/notifications"
    CACHE_PATH = Path(VPNExecutionEnvironment().path_cache) / "notifications.json"

    def __init__(
        self, session: "VPNSession",
        refresh_calculator: RefreshCalculator = None,
        cache_handler: CacheHandler = None
    ):
        """
        :param session: session used to retrieve notifications.
        """
        self._notifications = None
        self._session = session
        self._refresh_calculator = refresh_calculator or RefreshCalculator
        self._cache_file = cache_handler or CacheHandler(self.CACHE_PATH)

    def clear_cache(self):
        """Discards the cache, if existing."""
        self._notifications = None
        self._cache_file.remove()

    async def fetch(self) -> Notifications:
        """
        Fetches notifications from the REST API.
        :returns: the fetched notifications.
        """
        response = await rest_api_request(
            self._session,
            self.ROUTE,
        )
        response["ExpirationTime"] = self._refresh_calculator\
            .get_expiration_time(refresh_interval=REFRESH_INTERVAL)
        self._propagate_seen_state_to_new_snapshot(response)
        self._cache_file.save(response)
        self._notifications = Notifications(response)

        return self._notifications

    def load_from_cache(self) -> Notifications:
        """
        Loads notifications from the cache.
        :returns: the cached notifications, or an empty and expired Notifications
                  instance if no cache was found (triggering an immediate fetch).
        """
        cache = self._cache_file.load()
        self._notifications = \
            Notifications(cache) if cache else Notifications({"ExpirationTime": 0})
        return self._notifications

    def set_notification_seen(self, seen_notification_id: str):
        """Finds a notification matching the ID and marks it as seen"""
        if self._notifications is None:
            return

        for notification in self._notifications.data.get("Notifications", []):
            if notification_id := notification.get("NotificationID", None):
                if notification_id == seen_notification_id:
                    notification["seen"] = True

        self._cache_file.save(self._notifications.data)

    def _is_notification_seen(self, seen_notification_id: str):
        if self._notifications is None:
            return False

        for notification in self._notifications.data.get("Notifications", []):
            if notification_id := notification.get("NotificationID", None):
                if notification_id == seen_notification_id:
                    return notification.get("seen", False)

        return False

    def _propagate_seen_state_to_new_snapshot(self, new_notification_data: dict):
        for notification in new_notification_data.get("Notifications", []):
            if notification_id := notification.get("NotificationID", None):
                if self._is_notification_seen(notification_id):
                    notification["seen"] = True
